use std::{
    error::Error,
    path::{Path, PathBuf},
    thread::{sleep, spawn},
    time::Duration,
};

use log::{debug, info};

use radicle::prelude::Profile;
use radicle_ci_broker::{
    adapter::Adapter, broker::Broker, config::Config, error::BrokerError, event::NodeEventSource,
    msg::Request, status::Status,
};

const STATUS_UPDATE_DELAY: Duration = Duration::from_millis(1000);

fn main() {
    if let Err(e) = fallible_main() {
        eprintln!("ERROR: {}", e);
        let mut e = e.source();
        while let Some(source) = e {
            eprintln!("caused by: {}", source);
            e = source.source();
        }
    }
}

fn fallible_main() -> Result<(), BrokerError> {
    pretty_env_logger::init();
    info!("Radicle CI broker starts");

    let mut args = std::env::args().skip(1);
    let filename: PathBuf = if let Some(filename) = args.next() {
        PathBuf::from(filename)
    } else {
        return Err(BrokerError::Usage);
    };

    let config = Config::load(&filename)?;
    debug!("loaded configuration: {:#?}", config);

    let mut broker = Broker::default();
    debug!("created broker");

    // FIXME: this is broken. the config file only lists how to invoke
    // each adapter, not what adapter to use for each repo.
    // for (rid, spec) in config.adapters.iter() {
    //     debug!("setting adapter for {rid:?} to {spec:#?}");
    //     let rid = RepoId::from_urn(rid).map_err(|e| BrokerError::BadRepoId(rid.into(), e))?;
    //     let adapter = Adapter::new(&spec.command).with_environment(spec.envs());
    //     broker.set_repository_adapter(&rid, &adapter);
    // }
    // debug!("set per-repository adapters");

    let spec =
        config
            .adapter(&config.default_adapter)
            .ok_or(BrokerError::UnknownDefaultAdapter(
                config.default_adapter.clone(),
            ))?;
    let adapter = Adapter::new(&spec.command).with_environment(spec.envs());
    broker.set_default_adapter(&adapter);
    debug!("set default adapter");

    let profile = Profile::load()?;
    debug!("loaded profile");

    let mut source = NodeEventSource::new(&profile)?;
    debug!("created node event source");

    for filter in config.filters.iter() {
        source.allow(filter.clone());
    }
    debug!("added filters to node event source");

    // Spawn a thread that updates the status page.
    let mut status = Status::new(config.status_page().unwrap_or(Path::new("/dev/null")));
    let s2 = status.clone();
    let _status_thread = spawn(move || status_updater(s2));

    // This loop ends when there's an error, e.g., failure to read an
    // event from the node.
    loop {
        debug!("waiting for event from node");
        for e in source.event()? {
            status.broker_event(&e);
            debug!("broker event {e:#?}");
            let req = Request::trigger(&profile, &e)?;
            broker.execute_ci(&req, &mut status)?;
        }
    }
}

fn status_updater(mut status: Status) {
    loop {
        if let Err(e) = status.write() {
            eprintln!("ERROR: failed to update status page: {e}");
        }
        sleep(STATUS_UPDATE_DELAY);
    }
}
